@@ -15,17 +15,39 @@ import logging
 
 
 class EqOdooConnection(OdooConnection):
-    def __init__(self, clean_old_reports, language, *args, **kwargs):
+    def __init__(self, clean_old_reports, language, collect_yaml, disable_qweb, workflow, *args, **kwargs):
         super(EqOdooConnection, self).__init__(*args, **kwargs)
         self.do_clean_reports = clean_old_reports
         self.language = language
+        self.collect_yaml = collect_yaml
+        self.disable_qweb = disable_qweb
+        self.workflow = workflow
 
-    def _search_report(self, model_name, report_name: dict):
+    def _search_report_v13(self, model_name, report_name: dict, IR_ACTIONS_REPORT=False, company_id=False):
         """
             If parameter "do_clean_reports" (delete_old_reports in connection yaml) is set, delete all report with
             report_type = fast_report
         """
-        IR_ACTIONS_REPORT = self.connection.env['ir.actions.report']
+        if not IR_ACTIONS_REPORT:
+            IR_ACTIONS_REPORT = self.connection.env['ir.actions.report']
+        report_ids = IR_ACTIONS_REPORT.search(
+            [('model', '=ilike', model_name), '|', ('name', '=ilike', report_name['ger']),
+             ('name', '=ilike', report_name['ger'] + " " + "(PDF)"),'|', ('company_id', '=', company_id), ('company_id', '=', False)])
+        if len(report_ids) == 0 and 'eng' in report_name:
+            report_ids = IR_ACTIONS_REPORT.search([('model', '=', model_name), '|', ('name', '=', report_name['eng']),
+                                                   ('name', '=', report_name['eng'] + " " + "(PDF)"), '|', ('company_id', '=', company_id), ('company_id', '=', False)])
+        if len(report_ids) == 0:
+            return False
+        else:
+            return report_ids[0]
+
+    def _search_report(self, model_name, report_name: dict, IR_ACTIONS_REPORT=False):
+        """
+            If parameter "do_clean_reports" (delete_old_reports in connection yaml) is set, delete all report with
+            report_type = fast_report
+        """
+        if not IR_ACTIONS_REPORT:
+            IR_ACTIONS_REPORT = self.connection.env['ir.actions.report']
         report_ids = IR_ACTIONS_REPORT.search(
             [('model', '=ilike', model_name), '|', ('name', '=ilike', report_name['ger']),
              ('name', '=ilike', report_name['ger'] + " " + "(PDF)")])
@@ -42,10 +64,30 @@ class EqOdooConnection(OdooConnection):
             If parameter "do_clean_reports" (delete_old_reports in connection yaml) is set, delete all report with
             report_type = fast_report
         """
-        if self.do_clean_reports:
-            report_ids = self._get_fast_report_ids()
-            for report_id in report_ids:
-                self._delete_report(report_id)
+        original_company_yaml_user = self.connection.env.user.company_id.id
+        company_ids = self.connection.env.user.company_ids.ids if self.connection.env.user.company_ids else self.connection.env.user.company_id.ids
+        for company_id in company_ids:
+            self.connection.env.user.company_id = company_id
+            if self.do_clean_reports:
+                report_ids = self._get_fast_report_ids()
+                for report_id in report_ids:
+                    self._delete_report(report_id)
+        self.connection.env.user.company_id = original_company_yaml_user
+
+    def check_dependencies(self, dependencies):
+        """
+            Check if all dependencies (modules) are installed, if one isn't, return False
+            :param: dependencies: List of names of modules
+        """
+        not_installed_modules = []
+        if dependencies:
+            for dependency in dependencies:
+                dependency_installed = self.check_module(dependency)
+                if not dependency_installed:
+                    not_installed_modules.append(dependency)
+            if not_installed_modules:
+                return False, not_installed_modules
+        return True, not_installed_modules
 
     def map_reports(self, report_list: list):
         """
@@ -55,14 +97,24 @@ class EqOdooConnection(OdooConnection):
         IR_MODEL = self.connection.env['ir.model']
         IR_MODEL_FIELDS = self.connection.env['ir.model.fields']
         IR_ACTIONS_REPORT = self.connection.env['ir.actions.report']
+        original_company_yaml_user = IR_ACTIONS_REPORT.env.user.company_id
         for report in report_list:
             report.self_ensure()
             report._data_dictionary['name'] = report.entry_name[self.language]
-            dependencies_installed = self.check_dependencies(report._dependencies)
-            if not dependencies_installed:
+            dependencies_installed, not_installed_modules = self.check_dependencies(report._dependencies)
+            if not dependencies_installed and not_installed_modules:
                 print(f"!!! ******** DEPENDENCIES FOR {report.report_name} NOT INSTALLED ******** !!!")
+                for not_installed_module in not_installed_modules:
+                    print(f"!!! ******** MODULE - {not_installed_module} - NOT INSTALLED ******** !!!")
                 continue
-            report_id = self._search_report(report.model_name, report.entry_name)
+            if report.company_id:
+                IR_ACTIONS_REPORT.env.user.company_id = report.company_id[0]
+                if self.version == '13':
+                    report_id = self._search_report_v13(report.model_name, report.entry_name, IR_ACTIONS_REPORT, report.company_id[0])
+                else:
+                    report_id = self._search_report(report.model_name, report.entry_name, IR_ACTIONS_REPORT)
+            else:
+                report_id = self._search_report(report.model_name, report.entry_name, IR_ACTIONS_REPORT)
             if not report_id:
                 report_id = IR_ACTIONS_REPORT.create(report._data_dictionary)
                 report_object = IR_ACTIONS_REPORT.browse(report_id)
@@ -72,6 +124,7 @@ class EqOdooConnection(OdooConnection):
             # Add report to print menu
             report_object.create_action()
             print(f"!!! ******** START {report.report_name} ******** !!!")
+            IR_ACTIONS_REPORT.env.user.company_id = original_company_yaml_user
             try:
                 # Loop over all models in report fields dictionary
                 for model_name in report._fields:
@@ -87,9 +140,10 @@ class EqOdooConnection(OdooConnection):
                             field = IR_MODEL_FIELDS.browse(field_id)
                             # Insert the report_id
                             if report_object.id not in field.eq_report_ids.ids:
-                                report_ids = field.eq_report_ids.ids
-                                report_ids.append(report_object.id)
+                                report_ids = field.eq_report_ids.ids + [report_object.id]
                                 field.write({'eq_report_ids': [(6, 0, report_ids)]})
+                                # field.update({'eq_report_ids': [(6, 0, report_ids)]})
+                                # field.update({'eq_report_ids': [(4, report_object.id)]})
                 if report._calculated_fields:
                     for field, content in report._calculated_fields.items():
                         for function_name, parameter in content.items():
@@ -101,28 +155,57 @@ class EqOdooConnection(OdooConnection):
                 print(ex)
 
     def collect_all_report_entries(self, output_path):
-        IR_MODEL_FIELDS = self.connection.env['ir.model.fields']
-        all_report_field_ids = IR_MODEL_FIELDS.search([('eq_report_ids', '!=', False)])
+        company_ids = self.connection.env.user.company_ids.ids if self.connection.env.user.company_ids else self.connection.env.user.company_id.ids
+        report_name_id_combination = dict()
         data_dictionary = {}
-        print('Collect fields...')
-        # Progressbar...
-        with click.progressbar(range(len(all_report_field_ids))) as bar:
-            for field_id in all_report_field_ids:
-                # Get object
-                field_object = IR_MODEL_FIELDS.browse(field_id)
-                # Get attributes
-                report_action_ids = field_object.eq_report_ids.ids
-                model_id = field_object.model_id
-                model_name = model_id.model
-                field_name = field_object.name
-                # Add field to dictionary
-                for report_action_id in report_action_ids:
-                    data_dictionary = self.add_field_to_dictionary(data_dictionary, report_action_id, model_name,
-                                                                   field_name)
+        for company_id in company_ids:
+            # Change the current company in the env
+            self.connection.env.user.company_id = company_id
+            
+            if self.version == "10":
+                IR_ACTIONS_REPORT = self.connection.env['ir.actions.report.xml']
+            else:
+                IR_ACTIONS_REPORT = self.connection.env['ir.actions.report']
+            data_dictionary_keys = list(data_dictionary.keys())
+            report_ids = IR_ACTIONS_REPORT.search([('report_type', '=', 'fast_report'),('id', 'not in', data_dictionary_keys)])
+
+            IR_MODEL_FIELDS = self.connection.env['ir.model.fields']
+            all_report_field_ids = IR_MODEL_FIELDS.search([('eq_report_ids', 'in', report_ids), ('eq_report_ids', '!=', False)])
+
+            # Get current company name
+            company_name = self.connection.env['res.company'].browse(company_id).name
+            print('Collect fields for %s' %company_name)
+            
+            # Progressbar...
+            with click.progressbar(all_report_field_ids, length=len(all_report_field_ids)) as bar:
+                for field_id in bar:
+                    # Get object
+                    field_object = IR_MODEL_FIELDS.browse(field_id)
+                    # Get attributes
+                    report_action_ids = field_object.eq_report_ids.ids
+                    model_id = field_object.model_id
+                    model_name = model_id.model
+                    field_name = field_object.name
+
+
+                    # Add field to dictionary
+                    for report_action_id in report_action_ids:
+                        report_action_object = IR_ACTIONS_REPORT.browse(report_action_id)
+                        company_id = report_action_object.company_id.id if report_action_object.company_id else False
+                        if company_id:
+                            if report_action_object.report_name in report_name_id_combination and report_action_id != report_name_id_combination[report_action_object.report_name]:
+                                if 'company_id' in data_dictionary[report_name_id_combination[report_action_object.report_name]] and company_id not in data_dictionary[report_name_id_combination[report_action_object.report_name]]['company_id']:
+                                    data_dictionary[report_name_id_combination[report_action_object.report_name]]['company_id'].append(company_id)
+                                continue
+                            else:
+                                report_name_id_combination[report_action_object.report_name] = report_action_id
+                        data_dictionary = self.add_field_to_dictionary(data_dictionary, report_action_id, model_name,
+                                                                    field_name, company_id)
         for report_action_id, fields in data_dictionary.items():
             # Create report object
             eq_report_object = self.create_eq_report_object(report_action_id, fields)
             eq_yaml_data = eq_report_object.ensure_data_for_yaml()
+
             # Get timestamp
             now = datetime.now()
             date_now = now.strftime("%m_%d_%Y_%H_%M_%S")
@@ -130,13 +213,18 @@ class EqOdooConnection(OdooConnection):
             output_name = output_path + '/' + eq_report_object.report_name + '_' + date_now + '.yaml'
             self.write_yaml(output_name, eq_yaml_data)
 
-    def add_field_to_dictionary(self, data_dictionary, report_id, model_name, field_name):
+    def add_field_to_dictionary(self, data_dictionary, report_id, model_name, field_name, company_id):
         if report_id not in data_dictionary:
             data_dictionary[report_id] = {}
         if model_name not in data_dictionary[report_id]:
             data_dictionary[report_id][model_name] = [field_name]
-        else:
+        if field_name not in data_dictionary[report_id][model_name]:
             data_dictionary[report_id][model_name].append(field_name)
+        if company_id:
+            if 'company_id' in data_dictionary[report_id] and company_id not in data_dictionary[report_id]['company_id']:
+                data_dictionary[report_id]['company_id'].append(company_id)
+            else:
+                data_dictionary[report_id]['company_id'] = [company_id]
         return data_dictionary
 
     def _collect_calculated_fields(self, eq_calculated_field_objects):
@@ -158,48 +246,63 @@ class EqOdooConnection(OdooConnection):
             }
         return eq_calculated_field_dict
 
+    def _collect_modules_dependencies_list(self, list_of_models):
+        IR_MODELS = self.connection.env['ir.model']
+        dependencies = []
+        for model in list_of_models:
+            model_id = IR_MODELS.search([('model','=', model)])
+            model_obj = IR_MODELS.browse(model_id)
+            models_dependencies = model_obj.modules.replace(' ', '').split(',')
+            dependencies.extend(models_dependencies)
+        # using set()
+        # to remove duplicated 
+        # from list
+        return sorted(list(set(dependencies)))
+
     def create_eq_report_object(self, action_id, field_dictionary):
         if self.version == "10":
             IR_ACTIONS_REPORT = self.connection.env['ir.actions.report.xml']
         else:
             IR_ACTIONS_REPORT = self.connection.env['ir.actions.report']
+        if 'company_id' in field_dictionary:
+            self.connection.env.user.company_id = field_dictionary['company_id'][0]
         action_object = IR_ACTIONS_REPORT.browse(action_id)
         # Collect attributes
         name = {self.language: action_object.name}
+        if self.language != 'eng':
+            name['eng'] = action_object.with_context(lang='en_US').name
         report_name = action_object.report_name
         report_type = action_object.report_type
         eq_export_type = action_object.eq_export_type
         print_report_name = action_object.print_report_name
         model_name = action_object.model
         eq_ignore_images = action_object.eq_ignore_images
-        eq_ignore_html = action_object.eq_ignore_html
-        eq_export_complete_html = action_object.eq_export_complete_html
-        multi_print = action_object.multi
+        eq_handling_html_fields = action_object.eq_handling_html_fields
+        multi = action_object.multi
         attachment_use = action_object.attachment_use
         attachment = action_object.attachment
         eq_calculated_field_ids = action_object.eq_calculated_field_ids
-        company_id = action_object.company_id
+        company_id = field_dictionary['company_id'] if 'company_id' in field_dictionary else False
+        if 'company_id' in field_dictionary:
+            del field_dictionary['company_id']
 
-        # Special cases => not in every version
-        eq_export_as_sql = action_object.eq_export_as_sql
-        if not self.is_boolean(eq_export_as_sql):
-            eq_export_as_sql = True
         calculated_fields_dict = self._collect_calculated_fields(eq_calculated_field_ids)
         if not self.is_dict(calculated_fields_dict):
             calculated_fields_dict = {}
         eq_print_button = action_object.eq_print_button
         if not self.is_boolean(eq_print_button):
             eq_print_button = False
-        eq_merge_data_from_multi = action_object.eq_merge_data_from_multi
-        if not self.is_boolean(eq_merge_data_from_multi):
-            eq_merge_data_from_multi = False
+        eq_multiprint = action_object.eq_multiprint
+        
+        list_of_models = list(field_dictionary.keys())
+        dependencies = self._collect_modules_dependencies_list(list_of_models)
 
         eq_report_obj = eq_report.EqReport(name, report_name, report_type, model_name, company_id, eq_export_type,
                                            print_report_name, attachment,
-                                           eq_ignore_images, eq_ignore_html, eq_export_complete_html, eq_export_as_sql,
-                                           multi_print,
-                                           attachment_use, eq_print_button, [], field_dictionary,
-                                           calculated_fields_dict, eq_merge_data_from_multi)
+                                           eq_ignore_images, eq_handling_html_fields,
+                                           multi,
+                                           attachment_use, eq_print_button, dependencies, field_dictionary,
+                                           calculated_fields_dict, eq_multiprint)
         return eq_report_obj
 
     def write_yaml(self, file_name, data):
@@ -229,12 +332,21 @@ class EqOdooConnection(OdooConnection):
         #logging.basicConfig(format='%(asctime)s - %(message)s ',  datefmt='%H:%M:%S %d.%m.%Y', level=logging.INFO, filename=f"testing_fast_report_rendering_{self.connection._env._db}.log")
         ## Console
         logging.basicConfig(format='%(asctime)s - %(message)s ',  datefmt='%H:%M:%S %d.%m.%Y', level=logging.INFO)
-
-        IR_ACTIONS_REPORT = self.connection.env['ir.actions.report']
-        IR_MODEL = self.connection.env['ir.model']
+        original_company_yaml_user = self.connection.env.user.company_id
         for report in report_list:
+            if report.company_id:
+                self.connection.env.user.company_id = report.company_id[0]
+                IR_ACTIONS_REPORT = self.connection.env['ir.actions.report']
+                IR_MODEL = self.connection.env['ir.model']
+                if self.version == '13':
+                    report_id = self._search_report_v13(report.model_name, report.entry_name, IR_ACTIONS_REPORT, report.company_id[0])
+                else:
+                    report_id = self._search_report(report.model_name, report.entry_name, IR_ACTIONS_REPORT)
+            else:
+                IR_ACTIONS_REPORT = self.connection.env['ir.actions.report']
+                IR_MODEL = self.connection.env['ir.model']
+                report_id = self._search_report(report.model_name, report.entry_name, IR_ACTIONS_REPORT)
             # Get report action record
-            report_id = self._search_report(report.model_name, report.entry_name)
             report_object = IR_ACTIONS_REPORT.browse(report_id) if report_id else False
             # Check if the report has been created and is type Fast Report
             if not report_id or report_object.report_type != "fast_report": 
@@ -277,3 +389,14 @@ class EqOdooConnection(OdooConnection):
                         logging.info(f"\033[0;31m!!! ******** REPORT \033[1;31m{report.report_name}\033[0;31m NOT RENDERING CORRECTLY ******** !!!\033[0;37m ")
                         logging.info(f"\033[0;31m!!! ******** EXCEPTION ******** !!!\033[0;37m")
                         logging.info("\033[0;31m" + str(ex) + "\033[0;37m")
+        self.connection.env.user.company_id = original_company_yaml_user
+
+    def disable_qweb_reports(self):
+        IR_ACTIONS_REPORT = self.connection.env['ir.actions.report']
+        report_ids = IR_ACTIONS_REPORT.search(
+            ['|', ('report_type', '=', 'qweb-pdf'), '|', ('report_type', '=', 'qweb-html'),
+             ('report_type', '=', 'qweb-text')])
+        for report_id in report_ids:
+            report_object = IR_ACTIONS_REPORT.browse(report_id)
+            report_object.unlink_action()
+        print("Disabled QWeb for " + self.database)
